import operator
from functools import reduce
from typing import Any, Callable, Dict, List, TypeVar

import numpy as np
import react_ipywidgets as react

from solara.hooks.misc import use_force_update, use_unique_key

max_unique = 100
T = TypeVar("T")

__all__ = [
    "df_type",
    "df_unique",
    "provide_cross_filter",
    "use_cross_filter",
    "use_df_pivot_data",
    "use_df_column_names",
    "max_unique",
]


def df_type(df):
    return df.__class__.__module__.split(".")[0]


class CrossFilterStore:
    def __init__(self) -> None:
        self.listeners: List[Callable] = []
        self.filters: Dict[str, Any] = {}

    def add(self, key, filter):
        self.filters[key] = filter

    def use(self, key):
        # we use this state to trigger update, we could do without
        updater = use_force_update()

        filter, set_filter = react.use_state(self.filters.get(key))

        def on_change():
            set_filter(self.filters.get(key))
            # even if we don't change our own filter, the other may change
            updater()

        def connect():
            self.listeners.append(on_change)

            def cleanup():
                self.listeners.remove(on_change)
                # also remove our filter, and notify the rest
                del self.filters[key]
                for listener in self.listeners:
                    listener()

            return cleanup

        react.use_effect(connect, [key])

        def setter(filter):
            self.filters[key] = filter
            for listener in self.listeners:
                listener()

        otherfilters = [filter for key_other, filter in self.filters.items() if key != key_other]
        return filter, otherfilters, setter


cross_filter_context = react.create_context(CrossFilterStore())


def provide_cross_filter():
    # create it once
    cross_filter_object = react.use_memo(CrossFilterStore)()
    cross_filter_context.provide(cross_filter_object)


def df_unique(df, column, limit=None):
    if df_type(df) == "vaex":
        return df.unique(column, limit=max_unique + 1, limit_raise=False)
    if df_type(df) == "pandas":
        x = df[column].unique()  # .to_numpy()
        return x[:limit]
    else:
        raise TypeError(f"{type(df)} not supported")


def use_df_column_names(df):
    if df_type(df) == "vaex":
        return df.get_column_names()
    elif df_type(df) == "pandas":
        return df.columns.tolist()
    else:
        raise TypeError(f"{type(df)} not supported")


def use_cross_filter(name, reducer: Callable[[T, T], T] = operator.and_):
    """Provides cross filtering, all other filters are combined using the reducer.

    Cross filtering will collect a set of filters (from other components), and combine
    them into a single filter, that excludes the filter we set for the current component.
    This is often used in dashboards where a filter is defined in a visualization component,
    but only applied to all other components.
    """
    key = use_unique_key(prefix=f"cross-filter-{name}-")
    cross_filter_store = react.use_context(cross_filter_context)
    _own_filter, otherfilters, set_filter = cross_filter_store.use(key)
    if otherfilters:
        cross_filter = reduce(reducer, otherfilters[1:], otherfilters[0])
    else:
        cross_filter = None
    return cross_filter, set_filter


def use_df_pivot_data(df, x, y, agg):
    agg_name = str(agg)
    data = {
        "x": x,
        "y": y,
        "agg": agg_name,
        "values": None,
        "values_x": None,
        "values_y": None,
        "headers": {"x": [], "y": []},
        "counts": {"x": 0, "y": 0},
    }

    columns = [*data["x"], *data["y"]]
    agg_column = "agg_count__"
    row_index = "__row_index__"

    def df_groupby_agg(df, columns, aggregate):
        if df_type(df) == "pandas":
            gb = df.groupby(columns, as_index=False)
            aggs = {agg_column: (columns[0], "size")}
            print(columns, aggs)
            return gb.agg(**aggs)
        elif df_type(df) == "vaex":
            return df.groupby(columns, sort=True).agg({agg_column: agg})
        else:
            raise TypeError(f"{type(df)} not supported")

    if data["x"] and data["y"]:
        dfg = df_groupby_agg(df, columns, agg)
    #         print(dfg)
    if data["x"]:
        dfgx = df_groupby_agg(df, data["x"], agg)
        #         dfgx = df.groupby(data['x'], sort=True).agg({agg_column: agg})
        dfgx[row_index] = np.arange(len(dfgx), dtype="int64")
    #         print('x', dfgx)
    #         print(dfgx[agg_column].tolist())
    #         dfgx[row_index] = vaex.vrange(0, len(dfgx), dtype='int64')
    if data["y"]:
        dfgy = df_groupby_agg(df, data["y"], agg)
        #         dfgx = df.groupby(data['x'], sort=True).agg({agg_column: agg})
        dfgy[row_index] = np.arange(len(dfgy), dtype="int64")
    #         print('y', dfgy)
    #         dfgx[row_index] = vaex.vrange(0, len(dfgx), dtype='int64')
    #         dfgy = df.groupby(data['y'], sort=True).agg({agg_column: agg})
    #         dfgy[row_index] = vaex.vrange(0, len(dfgy), dtype='int64')

    if df_type(df) == "pandas":
        #         dfg_total = df.agg(agg)
        dfg_total = df.agg(**{agg_column: (columns[0], "size")}).to_numpy()[0][0]
    elif df_type(df) == "vaex":
        # agg = vaex.agg.count()
        dfg_total = df._agg(agg)
    else:
        raise TypeError(f"{type(df)} not supported")

    def formatter(value):
        if isinstance(value, float):
            return f"{value:,.2f}"
        else:
            return f"{value:,d}"

    if data["x"] and data["y"]:
        data["values"] = [[None] * len(dfgy) for _ in range(len(dfgx))]
    if data["x"]:
        #         print(dfgx[agg_column].tolist())
        data["values_x"] = list(map(formatter, dfgx[agg_column].tolist()))
    if data["y"]:
        data["values_y"] = list(map(formatter, dfgy[agg_column].tolist()))
    data["total"] = formatter(dfg_total)
    #     print(data)

    if data["x"] and data["y"]:
        for row in dfg.to_records():
            dfs = {"x": dfgx, "y": dfgy}
            for column in columns:
                axis_name = "x" if column in data["x"] else "y"
                df_axis = dfs[axis_name]
                #                 print("before", df_axis)
                if row[column] is None:
                    df_axis = df_axis[df[column].ismissing()]
                else:
                    df_axis = df_axis[df_axis[column] == row[column]]
                #                 print("after", df_axis, column, row[column])
                dfs[axis_name] = df_axis
            assert len(dfs["x"])
            assert len(dfs["y"])
            xi = dfs["x"][row_index].tolist()[0]
            yi = dfs["y"][row_index].tolist()[0]
            value = row[agg_column]
            data["values"][xi][yi] = formatter(value)

    if data["x"]:
        data["headers"]["x"] = [dfgx[k].tolist() for k in data["x"]]
        data["counts"]["x"] = len(dfgx)
    if data["y"]:
        data["headers"]["y"] = [dfgy[k].tolist() for k in data["y"]]
        data["counts"]["y"] = len(dfgy)

    return data



"""
# Slider

To support proper typechecks, we have multiple slider (all wrapping the ipyvuetify slider).
"""
import datetime

from solara.kitchensink import react, sol


@react.component
def SliderDemo():
    with sol.VBox() as main:
        with sol.Card("Integers"):
            int_value, set_int_value = react.use_state(42)
            sol.IntSlider("Some integer", value=int_value, min=-10, max=120, on_value=set_int_value)
            sol.Markdown(f"**Int value**: {int_value}")

        with sol.Card("Floats"):
            float_value, set_float_value = react.use_state(42.4)
            sol.FloatSlider("Some float", value=float_value, min=-10, max=120, on_value=set_float_value)
            sol.Markdown(f"**Float value**: {float_value}")

        with sol.Card("Values"):
            values = "Python C++ Java JavaScript TypeScript BASIC".split()
            value, set_value = react.use_state(values[0])
            print("value", value)
            sol.ValueSlider("Language", value, values=values, on_value=set_value)
            sol.Markdown(f"**Value**: {value}")

        with sol.Card("Dates"):
            date, set_date = react.use_state(datetime.date(1981, 7, 28))
            sol.DateSlider("Some date", value=date, on_value=set_date)
            sol.Markdown(f"**Date**: {date.strftime('%Y-%b-%d')}")

    return main


App = SliderDemo
app = App()



import base64
import os
import sys
import typing
from pathlib import Path

import PIL.Image

import solara as sol

if typing.TYPE_CHECKING:
    import numpy as np


def github_url(file):
    rel_path = os.path.relpath(file, Path(sol.__file__).parent.parent)
    github_url = sol.github_url + f"/blob/{sol.git_branch}/" + rel_path
    return github_url


def load_file_as_data_url(file_name, mime):
    with open(file_name, "rb") as f:
        data = f.read()
    return f"data:{mime};base64," + base64.b64encode(data).decode("utf-8")


def isinstanceof(object, spec: str):
    """Check if object is instance of type '<modulename>:<classname>'

    This can avoid a runtime dependency, since we do not need to import `modulename`.

    >>> import numpy as np
    >>> isinstanceof(np.arange(2), "numpy:ndarray")
    True
    """
    module_name, classname = spec.split(":")
    module = sys.modules.get(module_name)
    if module:
        cls = getattr(module, classname)
        return isinstance(object, cls)
    return False


def numpy_to_image(data: "np.ndarray", format="png"):
    import io

    if data.ndim == 3:
        if data.shape[2] == 3:
            im = PIL.Image.fromarray(data[::], "RGB")
        elif data.shape[2] == 4:
            im = PIL.Image.fromarray(data[::], "RGBA")
        else:
            raise ValueError(f"Expected last dimension to have 3 or 4 dimensions, total shape we got was {data.shape}")
        f = io.BytesIO()
        im.save(f, format)
        return f.getvalue()
    else:
        raise ValueError(f"Expected an image with 3 dimensions (height, width, channe), not {data.shape}")



import contextlib
import sys

import pytest

from solara.server import kernel, patch
from solara.server.app import AppContext


@pytest.fixture(scope="session")
def solara_patched():
    patch.patch()


@pytest.fixture(autouse=True)
def app_context(solara_patched):
    kernel_shared = kernel.Kernel()
    context = AppContext(id="1", kernel=kernel_shared, control_sockets=[], widgets={}, templates={})
    try:
        with context:
            yield context
    finally:
        with context:
            context.close()


@pytest.fixture
def extra_include_path():
    @contextlib.contextmanager
    def extra_include_path(path):
        sys.path.insert(0, str(path))
        try:
            yield
        finally:
            sys.path[:] = sys.path[1:]

    return extra_include_path



import dataclasses
import os

import ipyvuetify as v
import ipywidgets
import traitlets

from ..datatypes import CellAction, ColumnAction

__all__ = [
    "PivotTable",
    "DataTable",
    "VegaLite",
    "Navigator",
    "GridLayout",
    "HTML",
    "watch",
]


class PivotTable(v.VuetifyTemplate):
    template_file = os.path.realpath(os.path.join(os.path.dirname(__file__), "vue/pivottable.vue"))
    d = traitlets.Dict(default_value={"no": "data"}).tag(sync=True)
    selected = traitlets.Dict(default_value=[]).tag(sync=True)


def _ensure_dict(d):
    if dataclasses.is_dataclass(d):
        return dataclasses.asdict(d)
    return d


def _drop_keys_from_list_of_mappings(drop):
    def closure(list_of_dicts, widget):
        return [{k: v for k, v in _ensure_dict(d).items() if k not in drop} for d in list_of_dicts]

    return closure


class DataTable(v.VuetifyTemplate):
    template_file = os.path.realpath(os.path.join(os.path.dirname(__file__), "vue/datatable.vue"))

    total_length = traitlets.CInt().tag(sync=True)
    checked = traitlets.List([]).tag(sync=True)  # indices of which rows are selected
    column_actions = traitlets.List(trait=traitlets.Instance(ColumnAction), default_value=[]).tag(
        sync=True, to_json=_drop_keys_from_list_of_mappings(["on_click"])
    )
    _column_actions_callbacks = traitlets.List(trait=traitlets.Callable(), default_value=[])
    cell_actions = traitlets.List(trait=traitlets.Instance(CellAction), default_value=[]).tag(sync=True, to_json=_drop_keys_from_list_of_mappings(["on_click"]))
    _cell_actions_callbacks = traitlets.List(trait=traitlets.Callable(), default_value=[])
    items = traitlets.Any().tag(sync=True)  # the data, a list of dict
    headers = traitlets.Any().tag(sync=True)
    headers_selections = traitlets.Any().tag(sync=True)
    options = traitlets.Any().tag(sync=True)
    items_per_page = traitlets.CInt(11).tag(sync=True)
    selections = traitlets.Any([]).tag(sync=True)
    selection_colors = traitlets.Any([]).tag(sync=True)
    selection_enabled = traitlets.Bool(True).tag(sync=True)
    highlighted = traitlets.Int(None, allow_none=True).tag(sync=True)
    scrollable = traitlets.Bool(False).tag(sync=True)

    # for use with scrollable, when used in the default UI
    height = traitlets.Unicode(None, allow_none=True).tag(sync=True)

    hidden_components = traitlets.List([]).tag(sync=False)

    def vue_on_column_action(self, data):
        header_value, action_index = data
        on_click = self._column_actions_callbacks[action_index]
        if on_click:
            on_click(header_value)

    def vue_on_cell_action(self, data):
        row, header_value, action_index = data
        on_click = self._cell_actions_callbacks[action_index]
        if on_click:
            on_click(header_value, row)


class VegaLite(v.VuetifyTemplate):
    template_file = os.path.realpath(os.path.join(os.path.dirname(__file__), "vue/vegalite.vue"))
    spec = traitlets.Dict().tag(sync=True)
    listen_to_click = traitlets.Bool(False).tag(sync=True)
    listen_to_hover = traitlets.Bool(False).tag(sync=True)
    on_click = traitlets.traitlets.Callable(None, allow_none=True)
    on_hover = traitlets.traitlets.Callable(None, allow_none=True)

    def vue_altair_click(self, *args):
        if self.on_click:
            self.on_click(*args)

    def vue_altair_hover(self, *args):
        if self.on_hover:
            self.on_hover(*args)


class Navigator(v.VuetifyTemplate):
    template_file = os.path.realpath(os.path.join(os.path.dirname(__file__), "vue/navigator.vue"))
    location = traitlets.Unicode(None, allow_none=True).tag(sync=True)


class GridLayout(v.VuetifyTemplate):
    template_file = os.path.join(os.path.dirname(__file__), "vue/gridlayout.vue")
    gridlayout_loaded = traitlets.Bool(False).tag(sync=True)
    items = traitlets.Union([traitlets.List(), traitlets.Dict()], default_value=[]).tag(sync=True, **ipywidgets.widget_serialization)
    grid_layout = traitlets.List(default_value=[]).tag(sync=True)
    draggable = traitlets.CBool(True).tag(sync=True)
    resizable = traitlets.CBool(True).tag(sync=True)


class HTML(v.VuetifyTemplate):
    template_file = os.path.realpath(os.path.join(os.path.dirname(__file__), "vue/html.vue"))
    tag = traitlets.Unicode("div").tag(sync=True)
    attributes = traitlets.Dict().tag(sync=True)
    unsafe_innerHTML = traitlets.Unicode(None, allow_none=True).tag(sync=True)


def watch():
    import ipyvue

    ipyvue.watch(os.path.realpath(os.path.dirname(__file__) + "/vue"))



import contextlib
import dataclasses
import functools
import inspect
import io
import json
import logging
import os
import sys

# import tempfile
import threading
import time
import urllib.request
import uuid
from typing import IO, Any, Callable, Iterator, Optional, Tuple, TypeVar, Union, cast

import react_ipywidgets as react

from solara.datatypes import FileContentResult, Result, ResultState

logger = logging.getLogger("react-ipywidgets.extra.hooks")
chunk_size_default = 1024**2

__all__ = [
    "use_thread",
    "use_download",
    "use_fetch",
    "use_json_load",
    "use_json",
    "use_file_content",
    "use_uuid4",
    "use_unique_key",
    "use_state_or_update",
    "use_previous",
]
T = TypeVar("T")
U = TypeVar("U")

MaybeResult = Union[T, Result[T]]


# inherit from BaseException so less change of being caught
# in an except
class CancelledError(BaseException):
    pass


def use_retry(*actions: Callable[[], Any]):
    counter, set_counter = react.use_state(0)

    def retry():
        for action in actions:
            action()
        set_counter(lambda counter: counter + 1)

    return counter, retry


def use_thread(
    callback=Union[
        Callable[[threading.Event], T],
        Iterator[Callable[[threading.Event], T]],
        Callable[[], T],
        Iterator[Callable[[], T]],
    ],
    dependencies=[],
    intrusive_cancel=True,
) -> Result[T]:
    def make_event(*_ignore_dependencies):
        return threading.Event()

    def make_lock():
        return threading.Lock()

    lock: threading.Lock = react.use_memo(make_lock)()
    updater = use_force_update()
    result_state, set_result_state = react.use_state(ResultState.INITIAL)
    error = react.use_ref(cast(Optional[Exception], None))
    result = react.use_ref(cast(Optional[T], None))
    running_thread = react.use_ref(cast(Optional[threading.Thread], None))
    counter, retry = use_retry()
    cancel: threading.Event = react.use_memo(make_event)(dependencies + [counter])

    @contextlib.contextmanager
    def cancel_guard():
        if not intrusive_cancel:
            yield
            return

        def tracefunc(frame, event, arg):
            # this gets called at least for every line executed
            if cancel.is_set():
                rc = react.core._get_render_context(required=False)
                # we do not want to cancel the rendering cycle
                if rc is None or not rc._is_rendering:
                    # this will bubble up
                    raise CancelledError()
            # keep tracing:
            return tracefunc

        # see https://docs.python.org/3/library/sys.html#sys.settrace
        # it is for the calling thread only
        # not every Python implementation has it
        prev = None
        if hasattr(sys, "gettrace"):
            prev = sys.gettrace()
        if hasattr(sys, "settrace"):
            sys.settrace(tracefunc)
        try:
            yield
        finally:
            if hasattr(sys, "settrace"):
                sys.settrace(prev)

    def run():
        result.current = None
        set_result_state(ResultState.STARTING)

        def runner():
            wait_for_thread = None
            with lock:
                # if there is a current thread already, we'll need
                # to wait for it. copy the ref, and set ourselves
                # as the current one
                if running_thread.current:
                    wait_for_thread = running_thread.current
                running_thread.current = threading.current_thread()
            if wait_for_thread is not None:
                set_result_state(ResultState.WAITING)
                # don't start before the previous is stopped
                try:
                    wait_for_thread.join()
                except:  # noqa
                    pass
                if threading.current_thread() != running_thread.current:
                    # in case a new thread was started that also was waiting for the previous
                    # thread to st stop, we can finish this
                    return
            error.current = None
            result.current = None
            set_result_state(ResultState.RUNNING)

            sig = inspect.signature(callback)
            if sig.parameters:
                f = functools.partial(callback, cancel)
            else:
                f = callback
            try:
                try:
                    # we only use the cancel_guard context manager around
                    # the function calls to f. We don't want to guard around
                    # a call to react, since that might slow down rendering
                    # during rendering
                    with cancel_guard():
                        value = f()
                    if inspect.isgenerator(value):
                        while True:
                            try:
                                with cancel_guard():
                                    result.current = next(value)
                            except StopIteration:
                                break
                            # assigning to the ref doesn't trigger a rerender, so do it manually
                            updater()
                        if threading.current_thread() == running_thread.current:
                            set_result_state(ResultState.FINISHED)
                    else:
                        result.current = value
                        if threading.current_thread() == running_thread.current:
                            set_result_state(ResultState.FINISHED)
                except Exception as e:
                    error.current = e
                    if threading.current_thread() == running_thread.current:
                        set_result_state(ResultState.ERROR)
                        logger.exception(e)
                    return
                except CancelledError:
                    pass
                    # this means this thread is cancelled not be request, but because
                    # a new thread is running, we can ignore this
            finally:
                if threading.current_thread() == running_thread.current:
                    running_thread.current = None
                    logger.info("thread done!")
                    if cancel.is_set():
                        set_result_state(ResultState.CANCELLED)

        logger.info("starting thread: %r", runner)
        thread = threading.Thread(target=runner, daemon=True)
        thread.start()

        def cleanup():
            cancel.set()  # cleanup for use effect

        return cleanup

    react.use_side_effect(run, dependencies + [counter])
    return Result[T](value=result.current, error=error.current, state=result_state, cancel=cancel.set, _retry=retry)


def use_download(
    f: MaybeResult[Union[str, os.PathLike, IO]], url, expected_size=None, delay=None, return_content=False, chunk_size=chunk_size_default
) -> Result:
    if not isinstance(f, Result):
        f = Result(value=f)
    assert isinstance(f, Result)
    content_length, set_content_length = react.use_state(expected_size, key="content_length")
    downloaded_length = 0
    file_object = hasattr(f.value, "tell")
    if not file_object:
        file_path = cast(Union[str, os.PathLike], f.value)
        if os.path.exists(file_path) and expected_size is not None:
            file_size = os.path.getsize(file_path)
            if file_size == expected_size:
                downloaded_length = file_size
    downloaded_length, set_downloaded_length = react.use_state(downloaded_length, key="downloaded_length")

    def download(cancel: threading.Event):
        assert isinstance(f, Result)
        nonlocal downloaded_length
        if expected_size is not None and downloaded_length == expected_size:
            return  # we already downloaded, but hooks cannot be conditional

        context: Any = None
        if file_object:
            context = contextlib.nullcontext()
            output_file = cast(IO, f.value)
        else:
            # f = cast(Result[Union[str, os.PathLike]], f)
            output_file = context = open(f.value, "wb")  # type: ignore

        with context:
            with urllib.request.urlopen(url) as response:
                content_length = int(response.info()["Content-Length"])
                logger.info("content_length for %r = %r", url, content_length)
                set_content_length(content_length)
                bytes_read = 0
                while not cancel.is_set():
                    chunk = response.read(chunk_size)

                    if delay:
                        time.sleep(delay)
                    if not chunk:
                        break
                    bytes_read += len(chunk)
                    output_file.write(chunk)
                    set_downloaded_length(bytes_read)
        return bytes_read

    result: Result[Any] = use_thread(download, [f, url])
    # maybe we wanna check this
    # download_is_done = downloaded_length == content_length

    if content_length is not None:
        progress = downloaded_length / content_length
    else:
        progress = 0

    return dataclasses.replace(result, progress=progress)


def use_fetch(url, chunk_size=chunk_size_default):
    # re-use the same file like object
    f = react.use_memo(lambda *ignore: io.BytesIO(), args=[url])
    result = use_download(f, url, return_content=True, chunk_size=chunk_size)
    print(result.progress)
    return dataclasses.replace(result, value=f.getvalue() if result.progress == 1 else None)


def compose_result(head, *tail):
    return head


def ensure_result(input: MaybeResult[T]) -> Result[T]:
    if isinstance(input, Result):
        return input
    else:
        return Result(value=input)


def make_use_thread(f: Callable[[T], U]):
    def use_result(input: MaybeResult[T]) -> Result[U]:
        input_result = ensure_result(input)

        def in_thread(cancel: threading.Event):
            if input_result.value:
                return f(input_result.value)

        return use_thread(in_thread, dependencies=[input_result.value])

    return use_result


@make_use_thread
def use_json_load(value: bytes):
    return json.loads(value)


def use_json(path):
    return use_fetch(path) | use_json_load


def use_file_content(path, watch=False) -> FileContentResult[bytes]:
    counter, retry = use_retry()

    @react.use_memo
    def read_file(*ignore):
        try:
            with open(path, "rb") as f:
                return f.read()
        except Exception as e:
            return e

    try:
        mtime = os.path.getmtime(path)
    except Exception as e:
        result = FileContentResult[bytes](error=e, _retry=retry)
        # result.retry = retry
        return result

    content = read_file(
        path,
        mtime,
        counter,
    )
    if isinstance(content, Exception):
        return FileContentResult[bytes](error=content, _retry=retry)
    else:
        return FileContentResult[bytes](value=content, _retry=retry)


def use_force_update() -> Callable[[], None]:
    """Returns a callable that can be used to force an update of a component.

    This is used when external state has change, and we need to re-render out component.
    """
    _counter, set_counter = react.use_state(0, "force update counter")

    def updater():
        set_counter(lambda count: count + 1)

    return updater


def use_uuid4(dependencies=[]):
    """Generate a unique string using the uuid4 algorithm. Will only change when the dependencies change."""

    @react.use_memo
    def make_uuid(*_ignore):
        return str(uuid.uuid4())

    return make_uuid(dependencies)


def use_unique_key(key: str = None, prefix: str = "", dependencies=[]):
    """Generate a unique string, or use key when not None. Dependencies are forwarded to `use_uuid4`."""
    uuid = use_uuid4(dependencies=dependencies)
    return prefix + (key or uuid)


def use_state_or_update(
    initial_or_updated: T, key: str = None, eq: Callable[[Any, Any], bool] = None
) -> Tuple[T, Callable[[Union[T, Callable[[T], T]]], None]]:
    """This is useful for situations where a prop can change from a parent
    component, which should be respected, and otherwise the internal
    state should be kept.
    """
    value, set_value = react.use_state(initial_or_updated, key=key, eq=eq)

    def possibly_update(initial_or_updated):
        nonlocal value
        # only gets called when initial_or_updated changes
        set_value(initial_or_updated)
        # this make sure the return value gets updated directly
        value = initial_or_updated

    react.use_memo(possibly_update)(initial_or_updated)
    return value, set_value


def use_previous(value: T) -> T:
    ref = react.use_ref(value)

    def assign():
        ref.current = value

    react.use_effect(assign, [value])
    return ref.current



import dataclasses
from enum import Enum
from typing import Any, Callable, Generic, Optional, TypeVar

T = TypeVar("T")
U = TypeVar("U")


@dataclasses.dataclass(frozen=True)
class Action:
    name: str
    icon: Optional[str] = None
    on_click: Optional[Callable] = None


@dataclasses.dataclass(frozen=True)
class ColumnAction(Action):
    on_click: Optional[Callable[[str], None]] = None


@dataclasses.dataclass(frozen=True)
class CellAction(Action):
    on_click: Optional[Callable[[str, int], None]] = None


def _fallback_retry():
    raise RuntimeError("Should not happen")


class ResultState(Enum):
    INITIAL = 1
    STARTING = 2
    WAITING = 3
    RUNNING = 4
    ERROR = 5
    FINISHED = 6
    CANCELLED = 7


@dataclasses.dataclass(frozen=True)
class Result(Generic[T]):
    value: Optional[T] = None
    error: Optional[Exception] = None
    state: ResultState = ResultState.INITIAL
    progress: Optional[float] = None

    def retry(self):
        # mypy does not like members that are callable
        # gets confused about self argument.
        # we wrap it to avoid hitting this error in user
        # code
        self._retry()  # type: ignore

    _retry: Callable[[], Any] = lambda: None
    cancel: Callable[[], Any] = lambda: None

    def __or__(self, next: Callable[["Result[T]"], "Result[U]"]):
        return next(self)


@dataclasses.dataclass(frozen=True)
class FileContentResult(Result[T]):
    @property
    def exists(self):
        return not isinstance(self.error, FileNotFoundError)



import atexit
import contextlib
import dataclasses
import importlib.util
import logging
import os
import pickle
import sys
import threading
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Optional, cast

import ipywidgets as widgets
import react_ipywidgets as react

from . import kernel, reload
from .utils import nested_get

WebSocket = Any


COOKIE_KEY_CONTEXT_ID = "solara-context-id"
thread_lock = threading.Lock()

logger = logging.getLogger("solara.server.app")
state_directory = Path(".") / "states"
state_directory.mkdir(exist_ok=True)


@contextlib.contextmanager
def cwd(path):
    cwd = os.getcwd()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(cwd)


@dataclasses.dataclass
class AppContext:
    id: str
    kernel: kernel.Kernel
    control_sockets: List[WebSocket]
    # this is the 'private' version of the normally global ipywidgets.Widgets.widget dict
    # see patch.py
    widgets: Dict[str, widgets.Widget]
    # same, for ipyvue templates
    # see patch.py
    templates: Dict[str, widgets.Widget]
    user_dicts: Dict[str, Dict] = dataclasses.field(default_factory=dict)
    # anything we need to attach to the context
    # e.g. for a react app the render context, so that we can store/restore the state
    app_object: Optional[Any] = None

    def display(self, *args):
        print(args)

    def __enter__(self):
        key = get_current_thread_key()
        current_context[key] = self

    def __exit__(self, *args):
        key = get_current_thread_key()
        current_context[key] = None

    def close(self):
        with self:
            import solara.server.patch

            assert isinstance(widgets.Widget.widgets, solara.server.patch.context_dict_widgets)
            assert widgets.Widget.widgets._get_context_dict() is self.widgets
            widgets.Widget.close_all()
            # what if we reference eachother
            # import gc
            # gc.collect()

    def state_reset(self):
        path = state_directory / f"{self.id}.pickle"
        path = path.absolute()
        try:
            path.unlink()
        except:  # noqa
            pass
        del contexts[self.id]
        key = get_current_thread_key()
        del current_context[key]

    def state_save(self, state_directory: os.PathLike):
        path = Path(state_directory) / f"{self.id}.pickle"
        render_context = self.app_object
        if render_context is not None:
            render_context = cast(react.core._RenderContext, render_context)
            state = render_context.state_get()
            with path.open("wb") as f:
                logger.debug("State: %r", state)
                pickle.dump(state, f)


contexts: Dict[str, AppContext] = {}
# maps from thread key to AppContext, if AppContext is None, it exists, but is not set as current
current_context: Dict[str, Optional[AppContext]] = {}


def get_current_thread_key() -> str:
    thread = threading.currentThread()
    return get_thread_key(thread)


def get_thread_key(thread: threading.Thread) -> str:
    thread_key = thread._name  # type: ignore
    return thread_key


def set_context_for_thread(context: AppContext, thread: threading.Thread):
    key = get_thread_key(thread)
    contexts[key] = context
    current_context[key] = context


def get_current_context() -> AppContext:
    thread_key = get_current_thread_key()
    if thread_key not in current_context:
        raise RuntimeError(
            f"Tried to get the current context for thread {thread_key}, but no known context found. This might be a bug in Solara. "
            f"(known contexts: {list(current_context.keys())}"
        )
    context = current_context[thread_key]
    if context is None:
        raise RuntimeError(
            f"Tried to get the current context for thread {thread_key}, although the context is know, it was not set for this thread. "
            + "This might be a bug in Solara."
        )
    return context


class AppType(str, Enum):
    SCRIPT = "script"
    NOTEBOOK = "notebook"
    MODULE = "module"


class AppScript:
    def __init__(self, name, default_app_name="app"):
        self.fullname = name
        if reload.reloader.on_change:
            raise RuntimeError("Previous reloader still had a on_change attached, no cleanup?")
        reload.reloader.on_change = self.on_file_change

        self.app_name = default_app_name
        if ":" in self.fullname:
            self.name, self.app_name = self.fullname.split(":")
        else:
            self.name = name
        self.path: Path = Path(self.name)
        if self.name.endswith(".py"):
            self.type = AppType.SCRIPT
            # manually add the script to the watcher
            reload.reloader.watcher.add_file(self.path)
        elif self.name.endswith(".ipynb"):
            self.type = AppType.NOTEBOOK
            # manually add the notebook to the watcher
            reload.reloader.watcher.add_file(self.path)
        else:
            # the module itself will be added by reloader
            # automatically
            with reload.reloader.watch():
                self.type = AppType.MODULE
                spec = importlib.util.find_spec(self.name)
                if spec is None:
                    raise ImportError(f"Module {self.name} cannot be found")
                assert spec is not None
                assert spec.origin is not None
                self.path = Path(spec.origin)
        # this is not expected for modules, similar to `python script.py and python -m package.mymodule`
        if self.type in [AppType.SCRIPT, AppType.NOTEBOOK]:
            working_directory = str(self.path.parent)
            if working_directory not in sys.path:
                sys.path.insert(0, working_directory)

        # this might be useful for development
        # but requires reloading of react in solara iself
        # for name, module in sys.modules.items():
        #     if name.startswith("react_ipywidgets"):
        #         file = inspect.getfile(module)
        #         self.watcher.add_file(file)

        # cgi vars: https://datatracker.ietf.org/doc/html/rfc3875
        os.environ["SCRIPT_NAME"] = self.name
        os.environ["PATH_TRANSLATED"] = str(self.path.resolve())

    def close(self):
        reload.reloader.on_change = None
        context_values = list(contexts.values())
        contexts.clear()
        for context in context_values:
            context.close()

    def run(self):
        with reload.reloader.watch():
            return self._run()

    def _run(self):
        context = get_current_context()
        local_scope = {"display": context.display, "__name__": "__main__", "__file__": str(self.path)}
        ignore = list(local_scope)
        if self.type == AppType.SCRIPT:
            with open(self.path) as f:
                ast = compile(f.read(), self.path, "exec")
                exec(ast, local_scope)
        elif self.type == AppType.NOTEBOOK:
            import nbformat

            nb: nbformat.NotebookNode = nbformat.read(self.path, 4)
            with cwd(Path(self.path).parent):
                for cell_index, cell in enumerate(nb.cells):
                    cell_index += 1  # used 1 based
                    if cell.cell_type == "code":
                        source = cell.source
                        cell_path = f"{self.path} input cell {cell_index}"
                        ast = compile(source, cell_path, "exec")
                        exec(ast, local_scope)
        elif self.type == AppType.MODULE:
            mod = importlib.import_module(self.name)
            local_scope = mod.__dict__
        else:
            raise ValueError(self.type)

        app = local_scope.get(self.app_name)
        app = nested_get(local_scope, self.app_name)
        if app is None:
            import difflib

            options = [k for k in list(local_scope) if k not in ignore and not k.startswith("_")]
            matches = difflib.get_close_matches(self.app_name, options)
            msg = f"No object with name {self.app_name} found for {self.name} at {self.path}."
            if matches:
                msg += " Did you mean: " + " or ".join(map(repr, matches))
            else:
                msg += " We did find: " + " or ".join(map(repr, options))
            raise NameError(msg)

        return app

    def on_file_change(self, name):
        path = Path(name)
        if path.suffix == ".vue":
            logger.info("Vue file changed: %s", name)
            template_content = path.read_text()
            for context in list(contexts.values()):
                with context:
                    for filepath, widget in context.templates.items():
                        if filepath == str(path):
                            widget.template = template_content
        else:
            logger.info("Reload requires due to change in module: %s", name)
            self.reload()

    def reload(self):
        # if multiple files change in a short time, we want to do this
        # not concurrently. Even better would be to do a debounce?
        with thread_lock:
            # first, we pickle, before we unload modules
            context_values = list(contexts.values())
            logger.debug("Saving state...")
            contexts.clear()
            for context in context_values:
                try:
                    context.state_save(state_directory=state_directory)
                except:  # noqa
                    logger.exception("Could not save state, will continue")
                context.close()

            def send_reload():
                reload = {
                    "type": "reload",
                    "reason": "app changed",
                }
                for context in context_values:
                    for socket in context.control_sockets:
                        try:
                            socket.send_json(reload)
                        except:  # noqa
                            pass  # ignore disconnected clients

            send_reload()


def state_store_all():
    logger.info("Storing context state to disk")
    for context in contexts.values():
        context.state_save(state_directory=state_directory)


def state_load(context_name: str):
    path = state_directory / f"{context_name}.pickle"
    if path.exists():
        try:
            with path.open("rb") as f:
                return pickle.load(f)
                # return json.load(f)
        except Exception:
            logger.exception("Failed to load state for context %s", context_name)
            raise


atexit.register(state_store_all)



import inspect
import urllib.parse

from solara.kitchensink import react, sol, v

from . import (
    button,
    datatable,
    file_browser,
    griddraggable,
    gridfixed,
    hbox,
    html,
    image,
    markdown,
    slider,
    sql_code,
    togglebuttons,
    use_previous,
    use_state_or_update,
    use_thread,
    vbox,
)

modules = {
    "Button": button,
    "Image": image,
    "Slider": slider,
    "ToggleButtons": togglebuttons,
    "DataTable": datatable,
    "GridDraggable": griddraggable,
    "GridFixed": gridfixed,
    "VBox": vbox,
    "HBox": hbox,
    "Markdown": markdown,
    "HTML": html,
    "SqlCode": sql_code,
    "use_thread": use_thread,
    "FileBrowser": file_browser,
    "use_state_or_update": use_state_or_update,
    "use_previous": use_previous,
}


@react.component
def API():
    selected, on_selected = react.use_state("Overview")
    with sol.HBox(grow=True) as main:
        with v.NavigationDrawer(right=False, width="min-content", v_model=True, permanent=True):
            with v.List(dense=True):
                with v.ListItemGroup(v_model=selected, on_v_model=on_selected):
                    sol.ListItem("Overview")
                    with sol.ListItem("Input", icon_name="mdi-chevron-left-box"):
                        sol.ListItem("Button")
                        sol.ListItem("Slider")
                        sol.ListItem("ToggleButtons")
                        sol.ListItem("FileBrowser")
                    with sol.ListItem("Output", icon_name="mdi-chevron-right-box"):
                        sol.ListItem("Markdown")
                        sol.ListItem("HTML")
                        sol.ListItem("Image")
                        sol.ListItem("Code")
                        sol.ListItem("SqlCode")
                    with sol.ListItem("Viz", icon_name="mdi-chart-histogram"):
                        sol.ListItem("FigurePlotly")
                        sol.ListItem("AltairChart")
                    with sol.ListItem("Layout", icon_name="mdi-page-layout-sidebar-left"):
                        sol.ListItem("HBox")
                        sol.ListItem("VBox")
                        sol.ListItem("GridDraggable")
                        sol.ListItem("GridFixed")
                        sol.ListItem("App")
                    with sol.ListItem("Data", icon_name="mdi-database"):
                        sol.ListItem("DataTable")
                    with sol.ListItem("Hooks", icon_name="mdi-hook"):
                        sol.ListItem("use_fetch")
                        sol.ListItem("use_json")
                        sol.ListItem("use_thread")
                        sol.ListItem("use_previous")
                        sol.ListItem("use_state_or_update")
                    with sol.ListItem("Types", icon_name="mdi-fingerprint"):
                        sol.ListItem("Action")
                        sol.ListItem("ColumnAction")

        # with v.Card(class_="d-flex", style_="flex-grow: 1", elevation=0) as main:
        with sol.HBox(grow=True):
            with sol.Padding(4):
                if selected in modules:
                    WithCode(modules[selected])

    return main


@react.component
def WithCode(module):
    component = module.App
    show_code, set_show_code = react.use_state(False)
    with v.Sheet() as main:
        with v.Dialog(v_model=show_code, on_v_model=set_show_code):
            with v.Sheet(class_="pa-4"):
                code = inspect.getsource(component.f)
                code = code.replace("```", "~~~")
                pre = ""
                sol.MarkdownIt(
                    f"""
```python
{pre}{code}
```
"""
                )
        # It renders code better
        sol.Markdown(module.__doc__ or "# no docs yet")
        sol.Markdown("# Example")
        sol.Button("Show code", icon_name="mdi-eye", on_click=lambda: set_show_code(True), class_="ma-4")
        code = inspect.getsource(module)

        code_quoted = urllib.parse.quote_plus(code)
        url = f"https://test.solara.dev/try?code={code_quoted}"
        sol.Button("Run on solara.dev", icon_name="mdi-pencil", href=url, target="_blank")
        component()
    return main


app = API()



import json
from typing import Union

import ipywidgets
import js

from . import app, patch, server
from .kernel import BytesWrap, Kernel, WebsocketStreamWrapper
from .websocket import WebsocketWrapper

context_id = "single"


class Websocket(WebsocketWrapper):
    def send_text(self, data: str) -> None:
        js.sendToPage(data)

    def send_bytes(self, data: bytes) -> None:
        js.sendToPage(data)

    def close(self) -> None:
        pass

    def receive(self) -> Union[str, bytes]:
        return b""


ws = Websocket()


def start():
    patch.patch()
    kernel = Kernel()
    kernel.shell_stream = WebsocketStreamWrapper(ws, "shell")
    kernel.control_stream = WebsocketStreamWrapper(ws, "control")
    context = app.contexts[context_id] = app.AppContext(id=context_id, kernel=kernel, control_sockets=[], widgets={}, templates={})
    app_state = None
    with context:
        ipywidgets.register_comm_target(kernel)
        widget, render_context = server.run_app(app_state)
        context.widgets["content"] = widget
    context.app_object = render_context
    model_id = context.widgets["content"].model_id
    kernel.session.websockets.add(ws)
    return model_id


async def processKernelMessage(msg):
    msg = json.loads(msg)
    context = app.contexts[context_id]
    kernel = context.kernel
    msg_serialized = kernel.session.serialize(msg)

    channel = msg["channel"]
    if channel == "shell":
        msg = [BytesWrap(k) for k in msg_serialized]
        # TODO: because we use await, we probably need to use a context
        # manager that sets the app context in a async context, not just thread context
        with context:
            await kernel.dispatch_shell(msg)
    return 42



import json
import logging
import os
import pdb
import sys
import time
import traceback
from pathlib import Path
from typing import List, Optional
from uuid import uuid4

import ipywidgets as widgets
import jinja2
import react_ipywidgets
from jupyter_core.paths import jupyter_config_path
from react_ipywidgets.core import Element, render

from . import app, reload, settings, websocket
from .app import AppContext, AppScript
from .kernel import BytesWrap, Kernel, WebsocketStreamWrapper

directory = Path(__file__).parent
template_name = "solara.html.j2"

jinja_loader = jinja2.FileSystemLoader(str(directory / "templates"))
jinja_env = jinja2.Environment(loader=jinja_loader, autoescape=True)
solara_app = AppScript(os.environ.get("SOLARA_APP", "solara.examples:app"))
logger = logging.getLogger("solara.server.server")
nbextensions_ignorelist = [
    "jupytext/index",
    "nbextensions_configurator/config_menu/main",
    "jupytext/index",
    "nbdime/index",
    "voila/extension",
    "contrib_nbextensions_help_item/main",
    "execute_time/ExecuteTime",
    "dominocode/extension",
]


async def app_loop(ws: websocket.WebsocketWrapper, context_id: Optional[str]):
    if context_id is None:
        logging.warning(f"no context id cookie set ({app.COOKIE_KEY_CONTEXT_ID})")
        # to avoid very fast reconnects (we are in a thread anyway)
        time.sleep(0.5)
        return
    context = app.contexts.get(context_id)
    if context is None:
        logging.warning("invalid context id: %r", context_id)
        # to avoid very fast reconnects (we are in a thread anyway)
        time.sleep(0.5)
        return

    kernel = context.kernel
    kernel.shell_stream = WebsocketStreamWrapper(ws, "shell")
    kernel.control_stream = WebsocketStreamWrapper(ws, "control")

    # should we use excepthook ?
    kernel.session.websockets.add(ws)
    if True:
        while True:
            try:
                message = ws.receive()
            except websocket.WebSocketDisconnect:
                logger.debug("Disconnected")
                return
            if isinstance(message, str):
                msg = json.loads(message)
            else:
                from jupyter_server.base.zmqhandlers import deserialize_binary_message

                msg = deserialize_binary_message(message)

            msg_serialized = kernel.session.serialize(msg)
            channel = msg["channel"]
            if channel == "shell":
                msg = [BytesWrap(k) for k in msg_serialized]
                with context:
                    await kernel.dispatch_shell(msg)
            else:
                print("unknown channel", msg["channel"])


def control_loop(ws: websocket.WebsocketWrapper, context_id: Optional[str]):
    if context_id is None:
        ws.send_json({"type": "reload", "reason": "no context id found in cookie"})
        ws.close()
        return
    context = app.contexts.get(context_id)
    if context is None:
        ws.send_json({"type": "reload", "reason": "context does not exist (server reload?)"})
    if context:
        app.contexts[context_id].control_sockets.append(ws)
    ok = True

    while ok:
        try:
            msg = ws.receive_json()
            if msg["type"] == "state_reset":
                logger.info(f"reset state for context {context_id}")
                if context:
                    context.state_reset()
                ws.send_json({"type": "reload", "reason": "context id does not exist (server reload?)"})
            else:
                logger.error("Unknown msg: {msg}")
        except Exception:
            context = app.contexts.get(context_id)
            if context:
                try:
                    context.control_sockets.remove(ws)
                except ValueError:
                    pass  # could be removed from kernel.py
            ok = False
            # make sure it is closed
            try:
                ws.close()
            except:  # noqa
                pass


def run_app(app_state):
    # app.signal_hook_install()
    main_object = solara_app.run()

    render_context = None

    if isinstance(main_object, widgets.Widget):
        return main_object, render_context
    elif isinstance(main_object, Element):
        # container = widgets.VBox()
        import ipyvuetify

        container = ipyvuetify.Html(tag="div")
        # container = ipyvuetify.Html(tag="div")
        # support older versions of react
        result = render(main_object, container, handle_error=False, initial_state=app_state)
        if isinstance(result, tuple):
            container, render_context = result
        else:
            render_context = result
        return container, render_context
    else:
        extra = ""
        dotted = []
        for key, value in vars(main_object).items():
            if isinstance(value, (Element, widgets.Widget)):
                dotted.append(f"{solara_app.app_name}.{key}")
        if dotted:
            extra = " We did find that sub objects that might work: " + ", ".join(dotted)
        raise ValueError(f"Main object (with name {solara_app.app_name} in {solara_app.path}) is not a Widget or Element, but {type(main_object)}." + extra)


def read_root(context_id: Optional[str], base_url: str = "", render_kwargs={}, use_nbextensions=True):
    # context_id = None
    if context_id is None or context_id not in app.contexts:
        kernel = Kernel()
        if context_id is None:
            context_id = str(uuid4())
        context = app.contexts[context_id] = AppContext(id=context_id, kernel=kernel, control_sockets=[], widgets={}, templates={})
        with context:
            widgets.register_comm_target(kernel)
            assert kernel is Kernel.instance()
        try:
            with context:
                with reload.reloader.watch():
                    while True:
                        # reloading might take in extra dependencies, so the reload happens first
                        if reload.reloader.requires_reload:
                            reload.reloader.reload()
                        # reload before starting app, because we may load state using pickle
                        # if we do that before reloading, the classes are not compatible:
                        # e.g.: _pickle.PicklingError: Can't pickle <class 'testapp.Clicks'>: it's not the same object as testapp.Clicks
                        try:
                            app_state = app.state_load(context_id)
                            logger.debug("Loaded state: %r", app_state)
                        except Exception:
                            app_state = None
                        try:
                            widget, render_context = run_app(app_state)
                        except Exception:
                            if settings.main.use_pdb:
                                logger.exception("Exception, will be handled by debugger")
                                pdb.post_mortem()
                            raise

                        if render_context:
                            context.app_object = render_context
                        if not reload.reloader.requires_reload:
                            break

        except react_ipywidgets.core.ComponentCreateError as e:
            from rich.console import Console

            console = Console(record=True)
            console.print(e.rich_traceback)
            error = console.export_html()
            widget = widgets.HTML(f"<pre>{error}</pre>")
            # raise
        except Exception as e:
            error = ""
            error = "".join(traceback.format_exception(None, e, e.__traceback__))
            print(error, file=sys.stdout, flush=True)
            # widget = widgets.Label(value="Error, see server logs")
            import html

            error = html.escape(error)
            with context:
                widget = widgets.HTML(f"<pre>{error}</pre>")
            # raise
        context.widgets["content"] = widget
    else:
        context = app.contexts[context_id]

    model_id = context.widgets["content"].model_id

    if use_nbextensions:
        nbextensions = get_nbextensions()
    else:
        nbextensions = []

    resources = {
        "theme": "light",
        "nbextensions": nbextensions,
    }
    template: jinja2.Template = jinja_env.get_template(template_name)
    render_settings = {
        "model_id": model_id,
        "base_url": base_url,
        "resources": resources,
        "theme": settings.theme.dict(),
        "production": settings.main.mode == "production",
        **render_kwargs,
    }
    logger.info("Render setting for template: %r", render_settings)
    response = template.render(**render_settings)
    return response, context_id


def find_prefixed_directory(path):
    prefixes = [sys.prefix, os.path.expanduser("~/.local")]
    for prefix in prefixes:
        directory = f"{prefix}{path}"
        if Path(directory).exists():
            return directory
    else:
        raise RuntimeError(f"{path} not found at prefixes: {prefixes}")


def get_nbextensions_directories() -> List[Path]:
    from jupyter_core.paths import jupyter_path

    all_nb_directories = jupyter_path("nbextensions")
    # FIXME: remove IPython nbextensions path after a migration period
    try:
        from IPython.paths import get_ipython_dir
    except ImportError:
        pass
    else:
        all_nb_directories.append(Path(get_ipython_dir()) / "nbextensions")
    return [Path(k) for k in all_nb_directories]


def get_nbextensions() -> List[str]:
    read_config_path = [os.path.join(p, "serverconfig") for p in jupyter_config_path()]
    read_config_path += [os.path.join(p, "nbconfig") for p in jupyter_config_path()]
    # import inline since we don't want this dep for pyiodide
    from jupyter_server.services.config import ConfigManager

    config_manager = ConfigManager(read_config_path=read_config_path)
    notebook_config = config_manager.get("notebook")
    # except for the widget extension itself, since Voilà has its own
    load_extensions = notebook_config.get("load_extensions", {})
    if "jupyter-js-widgets/extension" in load_extensions:
        load_extensions["jupyter-js-widgets/extension"] = False
    if "voila/extension" in load_extensions:
        load_extensions["voila/extension"] = False
    directories = get_nbextensions_directories()

    def exists(name):
        for directory in directories:
            if (directory / (name + ".js")).exists():
                return True
        logger.error(f"nbextension {name} not found")
        return False

    nbextensions = [name for name, enabled in load_extensions.items() if enabled and (name not in nbextensions_ignorelist) and exists(name)]
    return nbextensions


if "pyodide" not in sys.modules:
    nbextensions_directories = get_nbextensions_directories()
    voila_static = find_prefixed_directory("/share/jupyter/voila/templates/base/static")
    nbconvert_static = find_prefixed_directory("/share/jupyter/nbconvert/templates/lab/static")
    solara_static = Path(__file__).parent / "static"



import inspect
import urllib.parse

from solara.kitchensink import react, sol, v

from . import calculator, pokemon, sine

tabs = {
    "Sine": sine,
    "Calulator": calculator,
    "Pokemon": pokemon,
}


@react.component
def Demo():
    tab, set_tab = react.use_state(0, "tab")

    # md, set_md = use_state("")
    with v.Tabs(v_model=tab, on_v_model=set_tab, vertical=True) as main:
        for key in tabs:
            with v.Tab(children=[key]):
                pass
        module = list(tabs.values())[tab]
        with v.TabsItems(v_model=tab):
            with sol.HBox(grow=False):
                with sol.VBox(grow=False):
                    github_url = sol.util.github_url(module.__file__)
                    with sol.HBox():
                        sol.Button("View on GitHub", icon_name="mdi-git", href=github_url, class_="ma-2", target="_blank")
                        code = inspect.getsource(module)

                        code_quoted = urllib.parse.quote_plus(code)
                        url = f"https://test.solara.dev/try?code={code_quoted}"
                        sol.Button("Run on solara.dev", icon_name="mdi-pencil", href=url, class_="ma-2", target="_blank")

                    module.App()
    return main



"""
# Image

```python
@react.component
def Image(image: Union[str, Path, "np.ndarray"]):
    ...
```

Displays an image from a URL, Path or numpy data.
"""

from pathlib import Path

import numpy as np
import PIL.Image

from solara.kitchensink import react, sol


@react.component
def ImageDemo():

    image_path = Path(sol.__file__).parent.resolve() / "server/static/sun64.png"
    print(image_path, Path(sol.__file__), sol.__file__)
    image_url = "/static/sun64.png"
    image_ndarray = np.asarray(PIL.Image.open(image_path))

    with sol.VBox() as main:
        with sol.Card(title="As a path"):
            sol.Image(image_path)

        with sol.Card(title="As a URL"):
            sol.Image(image_url)

        with sol.Card(title="As NumPy array"):
            sol.Image(image_ndarray)

    return main


Component = sol.Image
App = ImageDemo
app = App()



"""
# GridFixed

Lays out children in a grid with a fixed number of columns.
"""

from solara.kitchensink import react, sol

from .common import ColorCard


@react.component
def GridFixedDemo():
    with sol.VBox() as main:
        colors = "green red orange brown yellow pink".split()
        with sol.GridFixed(columns=3):
            for color in colors:
                ColorCard(color, color)
    return main


Component = sol.GridFixed
App = GridFixedDemo



import asyncio
import logging

import flask
import simple_websocket
from flask import Blueprint, Flask, request, send_from_directory, url_for
from flask_sock import Sock

from . import app as appmod
from . import server, websocket

logger = logging.getLogger("solara.server.flask")
blueprint = Blueprint("blueprint-solara", __name__)
websocket_extension = Sock()


class WebsocketWrapper(websocket.WebsocketWrapper):
    ws: simple_websocket.Server

    def __init__(self, ws: simple_websocket.Server) -> None:
        self.ws = ws

    def close(self):
        self.ws.close()

    def send_text(self, data: str) -> None:
        self.ws.send(data)

    def send_bytes(self, data: bytes) -> None:
        self.ws.send(data)

    def receive(self):
        return self.ws.receive()


@blueprint.route("/jupyter/api/kernels/<id>")
async def kernels(id):
    return {"name": "lala", "id": "dsa"}


@websocket_extension.route("/jupyter/api/kernels/<id>/<name>")
def kernels_connection(ws: simple_websocket.Server, id: str, name: str):
    context_id = request.cookies.get(appmod.COOKIE_KEY_CONTEXT_ID)
    ws_wrapper = WebsocketWrapper(ws)
    asyncio.run(server.app_loop(ws_wrapper, context_id=context_id))


@websocket_extension.route("/solara/watchdog/", websocket=True)
def watchdog(ws: simple_websocket.Server):
    ws_wrapper = WebsocketWrapper(ws)
    context_id = request.cookies.get(appmod.COOKIE_KEY_CONTEXT_ID)
    server.control_loop(ws_wrapper, context_id)


@blueprint.route("/static/nbextensions/<dir>/<filename>")
def nbext(dir, filename):
    for directory in server.nbextensions_directories:
        file = directory / dir / filename
        if file.exists():
            return send_from_directory(directory / dir, filename)
    return flask.Response("not found", status=404)


@blueprint.route("/static/dist/<path:path>")
def serve_voila_static(path):
    return send_from_directory(server.voila_static, path)


@blueprint.route("/static/nbconvert/<path:path>")
def serve_nbconvert_static(path):
    return send_from_directory(server.nbconvert_static, path)


@blueprint.route("/static/<path:path>")
def serve_static(path):
    return send_from_directory(server.solara_static, path)


@blueprint.route("/", defaults={"path": ""})
@blueprint.route("/<path:path>")
async def read_root(path):
    base_url = url_for(".read_root")
    if base_url.endswith("/"):
        base_url = base_url[:-1]
    print(base_url)
    context_id = request.cookies.get(appmod.COOKIE_KEY_CONTEXT_ID)
    content, context_id = server.read_root(context_id, base_url=base_url)
    assert context_id is not None
    response = flask.Response(content, mimetype="text/html")
    response.set_cookie(appmod.COOKIE_KEY_CONTEXT_ID, value=context_id)
    return response


# using the blueprint and websocket blueprint makes it easier to integrate into other applications
websocket_extension.init_app(blueprint)
app = Flask(__name__)
app.register_blueprint(blueprint)


if __name__ == "__main__":
    from .patch import patch

    patch()
    app.run(debug=False, port=5002)



from typing import Dict, List

import ipyvue
import react_ipywidgets as react
import traitlets


class SqlCodeWidget(ipyvue.VueTemplate):
    template_file = (__file__, "sql_code.vue")

    label = traitlets.Unicode("").tag(sync=True)
    query = traitlets.Unicode(allow_none=True, default_value=None).tag(sync=True)
    tables = traitlets.Dict(allow_none=True, default_value=None).tag(sync=True)
    height = traitlets.Unicode("180px").tag(sync=True)


@react.component
def SqlCode(label="Query", query: str = None, tables: Dict[str, List[str]] = None, on_query=None, height="180px"):
    return SqlCodeWidget.element(label=label, query=query, tables=tables, on_query=on_query, height=height)



import contextlib
import logging
from pathlib import Path

import playwright.sync_api

from solara.server import reload

app_path = Path(__file__).parent / "testapp.py"

logger = logging.getLogger("solara-test.integration.reload_test")


@contextlib.contextmanager
def append(text):
    with app_path.open() as f:
        content = f.read()
    try:
        with app_path.open("w") as f:
            f.write(content)
            f.write(text)
        yield
    finally:
        with app_path.open("w") as f:
            f.write(content)


@contextlib.contextmanager
def replace(path, text):
    with path.open() as f:
        content = f.read()
    try:
        with path.open("w") as f:
            f.write(text)
        yield
    finally:
        with path.open("w") as f:
            f.write(content)


def test_reload_syntax_error(page: playwright.sync_api.Page, solara_server, solara_app, extra_include_path):
    with extra_include_path(app_path.parent), solara_app("testapp:app"):
        # use as module, otherwise pickle wil not work
        page.goto(solara_server.base_url)
        assert page.title() == "Hello from Solara ☀️"
        page.locator("text=Clicked 0 times").click()
        page.locator("text=Clicked 1 times").click()

        with append("\n$%#$%"):
            reload.reloader.reload_event_next.wait()
            # page.locator("text=Clicked 2 times").click()
            page.locator("text=SyntaxError").wait_for()
        reload.reloader.reload_event_next.wait()
        page.locator("text=Clicked 2 times").click()
        page.locator("text=Clicked 3 times").wait_for()


def test_reload_many(page: playwright.sync_api.Page, solara_server, solara_app, extra_include_path):
    with extra_include_path(app_path.parent), solara_app("testapp:app"):
        logger.info("test_reload_many:run app")
        # use as module, otherwise pickle wil not work
        page.goto(solara_server.base_url)
        assert page.title() == "Hello from Solara ☀️"
        page.locator("text=Clicked 0 times").click()
        page.locator("text=Clicked 1 times").click()

        logger.info("test_reload_many:Touch app 1st time")
        app_path.touch()
        reload.reloader.reload_event_next.wait()
        page.locator("text=Clicked 2 times").click()
        page.locator("text=Clicked 3 times").wait_for(state="visible")

        logger.info("test_reload_many:Touch app 2st time")
        app_path.touch()
        reload.reloader.reload_event_next.wait()
        page.locator("text=Clicked 3 times").click()
        page.locator("text=Clicked 4 times").wait_for(state="visible")


def test_reload_vue(page: playwright.sync_api.Page, solara_server, solara_app, extra_include_path):
    with extra_include_path(app_path.parent), solara_app("testapp:vue_test_app"):
        page.goto(solara_server.base_url)
        assert page.title() == "Hello from Solara ☀️"
        page.locator("text=foobar").wait_for()

        vuecode = """
<template>
  <div>RELOADED</div>
</template>
        """
        vuepath = Path(__file__).parent / "test.vue"
        with replace(vuepath, vuecode):
            page.locator("text=RELOADED").wait_for()
        page.locator("text=foobar").wait_for()



import os
from os.path import isfile, join
from pathlib import Path
from typing import Callable, List, Optional, Union, cast

import humanize
import ipyvuetify as vy
import react_ipywidgets as react
import traitlets

import solara as sol
from solara.components import Div


def list_dir(path, filter=filter):
    def mk_item(n):
        full_path = join(path, n)
        is_file = isfile(full_path)
        return {"name": n, "is_file": is_file, "size": humanize.naturalsize(os.stat(full_path).st_size) if is_file else None}

    files = [mk_item(k) for k in os.listdir(path) if not k.startswith(".") if filter(Path(path) / k)]
    sorted_files = sorted(files, key=lambda item: ("0" if item["is_file"] else "1") + item["name"].lower())
    return sorted_files


class FileListWidget(vy.VuetifyTemplate):
    template_file = (__file__, "file_list_widget.vue")

    files = traitlets.List().tag(sync=True)
    clicked = traitlets.Dict(allow_none=True, default_value=None).tag(sync=True)
    double_clicked = traitlets.Dict(allow_none=True, default_value=None).tag(sync=True)
    scroll_pos = traitlets.Int(allow_none=True).tag(sync=True)

    def test_click(self, path: Union[Path, str], double_click=False):
        """Simulate a click or double click at the Python side"""
        matches = [k for k in self.files if k["name"] == str(path)]
        if len(matches) == 0:
            names = [k["name"] for k in self.files]
            raise NameError(f"Could not find {path}, possible filenames: {names}")
        item = matches[0]
        if double_click:
            self.double_clicked = item
        else:
            self.clicked = item

    def __contains__(self, name):
        """Test if filename/directory name is in the current directory."""
        return name in [k["name"] for k in self.files]


@react.component
def FileBrowser(
    directory: Union[str, Path] = None,
    on_directory_change: Callable[[Path], None] = None,
    on_path_select: Callable[[Optional[Path]], None] = None,
    on_file_open: Callable[[Path], None] = None,
    filter: Callable[[Path], bool] = lambda x: True,
    on_file_name: Callable[[str], None] = None,
    start_directory=None,
    can_select=False,
):
    """
    on_file_name is deprecated, use on_file_open
    start_directory is deprecated, use directory
    """
    if start_directory is not None:
        directory = start_directory  # pragma: no cover
    if directory is None:
        directory = os.getcwd()  # pragma: no cover
    current_dir, set_current_dir = sol.use_state_or_update(str(directory))
    selected, set_selected = react.use_state(None)
    double_clicked, set_double_clicked = react.use_state(None)
    warning, set_warning = react.use_state(cast(Optional[str], None))
    scroll_pos_stack, set_scroll_pos_stack = react.use_state(cast(List[int], []))
    scroll_pos, set_scroll_pos = react.use_state(0)
    selected, set_selected = react.use_state(None)

    def change_dir(new_dir):
        if os.access(new_dir, os.R_OK):
            set_current_dir(new_dir)
            if on_directory_change:
                on_directory_change(Path(new_dir))
            set_warning(None)
            return True
        else:
            set_warning(f"[no read access to {new_dir}]")

    def on_item(item, double_click):
        if item is None:
            if can_select and on_path_select:
                on_path_select(None)
            return
        if item["name"] == "..":
            new_dir = current_dir[: current_dir.rfind(os.path.sep)]
            action_change_directory = (can_select and double_click) or (not can_select and not double_click)
            if action_change_directory and change_dir(new_dir):
                if scroll_pos_stack:
                    last_pos = scroll_pos_stack[-1]
                    set_scroll_pos_stack(scroll_pos_stack[:-1])
                    set_scroll_pos(last_pos)
                set_selected(None)
                set_double_clicked(None)
            if can_select and not double_click:
                if on_path_select:
                    on_path_select(Path(new_dir))
            return

        path = os.path.join(current_dir, item["name"])
        is_file = item["is_file"]
        if (can_select and double_click) or (not can_select and not double_click):
            if is_file:
                if on_file_open:
                    on_file_open(Path(path))
                if on_file_name is not None:
                    on_file_name(path)
            else:
                if change_dir(path):
                    set_scroll_pos_stack(scroll_pos_stack + [scroll_pos])
                    set_scroll_pos(0)
            set_selected(None)
            set_double_clicked(None)
        elif can_select and not double_click:
            if on_path_select:
                on_path_select(Path(path))
        else:  # not can_select and double_click is ignored
            raise RuntimeError("Combination should not happen")  # pragma: no cover

    def on_click(item):
        set_selected(item)
        on_item(item, False)

    def on_double_click(item):
        set_double_clicked(item)
        if can_select:
            on_item(item, True)
        # otherwise we can ignore it, single click will handle it

    with Div(class_="solara-file-browser") as main:
        Div(children=[current_dir])
        FileListWidget.element(
            files=[{"name": "..", "is_file": False}] + list_dir(current_dir, filter=filter),
            selected=selected,
            clicked=selected,
            on_clicked=on_click,
            double_clicked=double_clicked,
            on_double_clicked=on_double_click,
            scroll_pos=scroll_pos,
            on_scroll_pos=set_scroll_pos,
        ).key("FileList")
        if warning:
            Div(style_="font-weight: bold; color: red", children=[warning])

    return main



"""
# FileBrowser

Browse file (and directories) at the server side.

There are two modes possible

   * `can_select=False`
      * `on_file_open`: Triggered when **single** clicking a file or directoy.
      * `on_path_select`: Never triggered
      * `on_directory_change`: Triggered when clicking a directory
   * `can_select=True`
      * `on_file_open`: Triggered when **double** clicking a file or directoy.
      * `on_path_select`: Triggered when clicking a file or directoy
      * `on_directory_change`: Triggered when double clicking a directory

"""
from pathlib import Path
from typing import Optional, cast

from solara.kitchensink import react, sol


@react.component
def FileBrowserDemo():
    file, set_file = react.use_state(cast(Optional[Path], None))
    path, set_path = react.use_state(cast(Optional[Path], None))
    directory, set_directory = react.use_state(Path("~").expanduser())

    with sol.VBox() as main:
        can_select = sol.ui_checkbox("Enable select")

        # reset path and file when can_select changes
        @react.use_memo
        def reset_path(can_select):
            set_path(None)
            set_file(None)

        reset_path(can_select)
        sol.FileBrowser(directory, on_directory_change=set_directory, on_path_select=set_path, on_file_open=set_file, can_select=can_select)
        sol.Info(f"You are in directory: {directory}")
        sol.Info(f"You selected path: {path}")
        sol.Info(f"You opened file: {file}")
    return main


Component = sol.FileBrowser
App = FileBrowserDemo
app = App()



import contextlib
import logging
import sys
import threading
from pathlib import Path

import playwright
import playwright.sync_api
import react_ipywidgets as react
import react_ipywidgets.ipywidgets as w

from solara.kitchensink import v

logger = logging.getLogger("solara.server.test")
HERE = Path(__file__).parent


@contextlib.contextmanager
def screenshot_on_error(page, path):
    try:
        yield
    except:  # noqa: E722
        page.screenshot(path=path)
        print(f"Saved screenshot to {path}", file=sys.stderr)
        raise


def test_docs_basics(page: playwright.sync_api.Page, solara_server, solara_app):
    # with screenshot_on_error(page, 'tmp/test_docs_basics.png'):
    with solara_app("solara.examples"):
        page.goto(solara_server.base_url)
        assert page.title() == "Hello from Solara ☀️"
        page.locator('div[role="tab"]:has-text("Demo")').click()

        page.locator("text=Calculator").click()
        page.locator("text=+/-").wait_for()
        page.screenshot(path="tmp/screenshot_calculator.png")

        page.locator("text=Supported").click()
        page.locator("text=Bqplot").click()
        page.locator("text=Line color").wait_for()
        page.screenshot(path="tmp/screenshot_bqplot.png")

        page.locator("text=Plotly").click()
        page.locator("text=plotly express").wait_for()
        page.screenshot(path="tmp/screenshot_plotly.png")

        page.locator("text=Altair").click()
        page.locator("text=Altair is supported").wait_for()
        page.screenshot(path="tmp/screenshot_altair.png")

        # page.locator("text=Docs").click()
        # page.screenshot(path="tmp/screenshot_debug.png")
        # page.locator('div[role="tab"]:has-text("use_state")').wait_for()
        # page.screenshot(path="tmp/screenshot_docs.png")

        # page.locator('div[role="tab"]:has-text("use_state")').click()
        # page.locator("text=use_state can be used").wait_for()
        # page.screenshot(path="tmp/screenshot_use_state.png")

        # page.locator('div[role="tab"]:has-text("use_effect")').click()
        # page.locator("text=use_side_effect can be used").wait_for()
        # page.screenshot(path="tmp/screenshot_use_effect.png")


@react.component
def ClickButton():
    count, set_count = react.use_state(0)
    if not isinstance(count, int):
        print("oops, state issue?")
        count = 0
    # return w.Button()  # description=, on_click=set_count(count + 1))
    btn = v.Btn(children=[f"Clicked: {count}"])

    def on_click(*ignore):
        set_count(count + 1)

    v.use_event(btn, "click", on_click)
    return btn


click_button = ClickButton()


def test_multi_user(page: playwright.sync_api.Page, solara_server, solara_app, extra_include_path):
    with extra_include_path(HERE), solara_app("server_test:click_button"):
        page.goto(solara_server.base_url)
        assert page.title() == "Hello from Solara ☀️"
        page.screenshot(path="tmp/screenshot_test_click.png")

        # page.locator("text=Clicked: 0").click()


@react.component
def ThreadTest():
    label, set_label = react.use_state("initial")
    use_thread, set_use_thread = react.use_state(False)

    def from_thread():
        set_label("from thread")

    def start_thread():
        if use_thread:
            thread = threading.Thread(target=from_thread)
            thread.start()
            return thread.join

    react.use_side_effect(start_thread, [use_thread])
    # we need to trigger creating a new widget, to make sure we
    # invoke a solara.server.app.get_current_context
    if label == "initial":
        return w.Button(description=label, on_click=lambda: set_use_thread(True))
    else:
        return w.Label(value=label)


thread_test = ThreadTest()
# click_button = ThreadTest()


def test_from_thread(page: playwright.sync_api.Page, solara_server, solara_app, extra_include_path):
    with extra_include_path(HERE), solara_app("server_test:thread_test"):
        page.goto(solara_server.base_url)

        assert page.title() == "Hello from Solara ☀️"
        el = page.locator(".jupyter-widgets")
        assert el.text_content() == "initial"
        page.wait_for_timeout(500)
        el.click()
        page.locator("text=from thread").wait_for()


def test_state(page: playwright.sync_api.Page, solara_server, solara_app, extra_include_path):
    with extra_include_path(HERE), solara_app("server_test:click_button"):
        page.goto(solara_server.base_url)
        # with screenshot_on_error(page, "tmp/test_state.png"):
        assert page.title() == "Hello from Solara ☀️"
        page.locator("text=Clicked: 0").click()
        page.locator("text=Clicked: 1").click()
        # refresh...
        page.goto(solara_server.base_url)
        # and state should be restored
        page.locator("text=Clicked: 2").wait_for()
        # account button
        page.locator("button >> nth=1").click()
        # reset state button
        page.locator('[role="menuitem"]').click()
        # refresh manually
        page.wait_for_timeout(100)
        page.goto(solara_server.base_url)
        # and state should be restored
        page.locator("text=Clicked: 0").wait_for()


def test_from_thread_two_users(browser: playwright.sync_api.Browser, solara_server, solara_app, extra_include_path):
    with extra_include_path(HERE), solara_app("server_test:thread_test"):
        context1 = browser.new_context()
        page1 = context1.new_page()
        context2 = browser.new_context()
        page2 = context2.new_page()

        page1.goto(solara_server.base_url)

        assert page1.title() == "Hello from Solara ☀️"
        el1 = page1.locator(".jupyter-widgets")
        assert el1.text_content() == "initial"

        page2.goto(solara_server.base_url)
        assert page2.title() == "Hello from Solara ☀️"
        el2 = page2.locator(".jupyter-widgets")
        assert el2.text_content() == "initial"

        page1.wait_for_timeout(500)
        page1.wait_for_timeout(500)

        el1.click()
        page1.locator("text=from thread").wait_for()

        page2.wait_for_timeout(500)
        assert el2.text_content() == "initial"

        el2.click()
        page2.locator("text=from thread").wait_for()


# def test_two_clients(browser: playwright.sync_api.Browser):
#     context1 = browser.new_context()
#     page1 = context1.new_page()
#     context2 = browser.new_context()
#     page2 = context1.new_page()



import threading

import react_ipywidgets as react
import traitlets
from ipyvue import Template
from ipyvuetify.extra import FileInput
from ipywidgets import widget_serialization

import solara.hooks as hooks


class FileDropZone(FileInput):
    # override to narrow traitlet of FileInput
    template = traitlets.Instance(Template).tag(sync=True, **widget_serialization)
    template_file = (__file__, "file_drop.vue")
    items = traitlets.List(default_value=[]).tag(sync=True)
    label = traitlets.Unicode().tag(sync=True)


@react.component
def FileDrop(on_total_progress, on_file, label="Drop file here"):
    file_info, set_file_info = react.use_state(None)
    wired_files, set_wired_files = react.use_state(None)

    file_drop = FileDropZone.element(label=label, on_total_progress=on_total_progress, on_file_info=set_file_info)

    def wire_files():
        if not file_info:
            return

        real = react.get_widget(file_drop)

        # workaround for @observe being cleared
        real.version += 1
        real.reset_stats()

        set_wired_files(real.get_files())

    react.use_side_effect(wire_files, [file_info])

    def handle_file(cancel: threading.Event):
        if not wired_files:
            return

        on_file(wired_files[0])

    hooks.use_thread(handle_file, [wired_files])

    return file_drop



"""
# use_state_or_update

Like `react.use_state`, except it will update the value if the input value changes.

This is useful to give a component state *or* allow the state to be controlled by the parent component


"""
from solara.kitchensink import react, sol


@react.component
def SliderWithoutState(value: int):
    # Note that this is very bad practive, if value is an input change and the slider
    # can change it, this component should have a on_value callback to allow the parent
    # component to manage its state.
    # This component is only for demoing/understanding.
    return sol.IntSlider("value", value=value)


@react.component
def SliderWithState(value: int):
    # effectively, because of use_state, the value prop passed in is
    # only a default value.
    value, set_value = react.use_state(value)
    return sol.IntSlider("value", value=value, on_value=set_value)


@react.component
def SliderWithStateOrUpdate(value: int):
    value, set_value = sol.use_state_or_update(value)
    return sol.IntSlider("value", value=value, on_value=set_value)


@react.component
def Test():
    parent_value, set_parent_value = react.use_state(4)
    # used to force rerenders
    rerender_counter, set_rerender_counter = react.use_state(4)
    with sol.VBox() as main:
        with sol.Card("Parent value selection"):
            sol.Info("This slider value gets passed down to the child components")
            sol.IntSlider("parent value", value=parent_value, on_value=set_parent_value)
            sol.Button("Force redraw", on_click=lambda: set_rerender_counter(rerender_counter + 1))

        with sol.Card("Child without state"):
            sol.Info("This child will simply render the value passed into the argument, a redraw will reset it to its parent value.")
            SliderWithoutState(parent_value)

        with sol.Card("Child with state"):
            sol.Info("This child will not care about the value passed into the prop, it manages its own state.")
            SliderWithState(parent_value)

        with sol.Card("Child with state (or update)"):
            sol.Info("This child will update when the passes in a new value, but a redraw will not reset it.")
            SliderWithStateOrUpdate(parent_value)

        with sol.Card("Child with state + key"):
            sol.Info(
                "We can also use the `.key(...)` method to force the component to forget its state, this will however cause the widget to be re-created"
                "(a performance penalty)."
            )
            SliderWithState(parent_value).key(f"slider-{parent_value}")

    return main


App = Test
app = App()



import threading
from typing import Callable, TypeVar

import pandas as pd
import pandas_ta as ta  # noqa: F401
import plotly.graph_objects as go
import typing_extensions

from solara.datatypes import Result
from solara.hooks.misc import use_thread
from solara.kitchensink import react, sol, v

P = typing_extensions.ParamSpec("P")
T = TypeVar("T")


def make_use_thread(f: Callable[P, T]):
    def use_result(*args: P.args, **kwargs: P.kwargs) -> Result[T]:
        def in_thread(cancel: threading.Event):
            return f(*args, **kwargs)

        return use_thread(in_thread, dependencies=[args, kwargs])

    return use_result


@make_use_thread
def use_load_df(ticker, period, interval) -> pd.DataFrame:
    @sol.memoize()
    def load(ticker, period, interval):
        e = pd.DataFrame()
        df = e.ta.ticker(ticker, period=period, interval=interval, kind="info", lc_cols=True)
        print(df)
        if df is None:
            raise ValueError("Error loading, nothing returned")
        return df

    return load(ticker, period, interval)


periods = "1d 5d 1mo 3mo 6mo 1y 2y 5y 10y ytd max".split()
intervals = "1m 2m 5m 15m 30m 60m 90m 1h 1d 5d 1wk 1mo 3mo".split()
tickers = ["BTC-USD", "SPY"]


@react.component
def App():
    ticker, set_ticker = react.use_state("SPY")
    period, set_period = react.use_state("1y")
    interval, set_interval = react.use_state("1wk")
    df_result = use_load_df(ticker, period=period, interval=interval)

    with sol.Div() as main:
        v.Select(items=tickers, v_model=ticker, on_v_model=set_ticker, label="Ticker")
        v.Select(items=periods, v_model=period, on_v_model=set_period, label="Period")
        v.Select(items=intervals, v_model=interval, on_v_model=set_interval, label="Interval")
        ema_length = sol.ui_slider(value=8, min=2, max=100, description="Length (EMA)")

        if df_result.error:
            sol.Error(f"Error: {df_result.error}")
            sol.Button("Retry", on_click=df_result.retry)
        else:
            if df_result.value is not None:
                df = df_result.value
                ema = df.ta.ema(length=ema_length)
                fig = go.Figure(
                    data=[
                        go.Ohlc(
                            x=df.index,
                            open=df["open"],
                            high=df["high"],
                            low=df["low"],
                            close=df["close"],
                            name=ticker,
                        ),
                        go.Scatter(x=ema.index, y=ema.values, name=f"{ticker}_EMA_{ema_length}", mode="lines"),
                    ]
                )
                sol.FigurePlotly(fig, dependencies=[])
            else:
                with v.Text(children=["Loading data..."]):
                    v.ProgressCircular(indeterminate=True, class_="solara-progress")

    return main


app = App()



import ast
import dataclasses
import operator
from typing import Any, Optional

import solara.util
from solara.kitchensink import react, sol, v

DEBUG = False
operator_map = {
    "x": operator.mul,
    "/": operator.truediv,
    "+": operator.add,
    "-": operator.sub,
}


github_url = solara.util.github_url(__file__)


@dataclasses.dataclass(frozen=True)
class CalculatorState:
    input: str = ""
    output: str = ""
    left: float = 0
    right: Optional[float] = None
    operator: Any = operator.add
    error: str = ""


initial_state = CalculatorState()


def calculate(state: CalculatorState):
    result = state.operator(state.left, state.right)
    return dataclasses.replace(state, left=result)


def calculator_reducer(state: CalculatorState, action):
    action_type, payload = action
    if DEBUG:
        print("reducer", state, action_type, payload)
    state = dataclasses.replace(state, error="")

    if action_type == "digit":
        digit = payload
        input = state.input + digit
        return dataclasses.replace(state, input=input, output=input)
    elif action_type == "percent":
        if state.input:
            try:
                value = ast.literal_eval(state.input)
            except Exception as e:
                return dataclasses.replace(state, error=str(e))
            state = dataclasses.replace(state, right=value / 100)
            state = calculate(state)
            output = f"{value / 100:,}"
            return dataclasses.replace(state, output=output, input="")
        else:
            output = f"{state.left / 100:,}"
            return dataclasses.replace(state, left=state.left / 100, output=output)
    elif action_type == "negate":
        if state.input:
            input = state.output
            input = input[1:] if input[0] == "-" else "-" + input
            output = input
            return dataclasses.replace(state, input=input, output=output)
        else:
            output = f"{-state.left:,}"
            return dataclasses.replace(state, left=-state.left, output=output)
    elif action_type == "clear":
        return dataclasses.replace(state, input="", output="")
    elif action_type == "reset":
        return initial_state
    elif action_type == "calculate":
        if state.input:
            try:
                value = ast.literal_eval(state.input)
            except Exception as e:
                return dataclasses.replace(state, error=str(e))
            state = dataclasses.replace(state, right=value)
        state = calculate(state)
        output = f"{state.left:,}"
        state = dataclasses.replace(state, output=output, input="")
        return state
    elif action_type == "operator":
        if state.input:
            state = calculator_reducer(state, ("calculate", None))
            state = dataclasses.replace(state, operator=payload, input="")
        else:
            # e.g. 2+3=*= should give 5,25
            state = dataclasses.replace(state, operator=payload, right=state.left)
        return state
    else:
        print("invalid action", action)
        return state


@react.component
def Calculator():
    state, dispatch = react.use_reducer(calculator_reducer, initial_state)
    if DEBUG:
        print("->", state)
    with v.Card(elevation=10, class_="ma-4") as main:
        with v.CardTitle(children=["Calculator"]):
            pass
        with v.CardSubtitle(children=["With ipyvuetify and ipywidgets-react"]):
            pass
        with v.CardText():
            with sol.VBox(grow=False):
                # with v.Container(style_="padding: 10px"):
                v.Label(children=[state.error or state.output or "0"])
                class_ = "pa-0 ma-1"

                with sol.HBox(grow=False):
                    if state.input:
                        v.BtnWithClick(children="C", on_click=lambda: dispatch(("clear", None)), dark=True, class_=class_)
                    else:
                        v.BtnWithClick(children="AC", on_click=lambda: dispatch(("reset", None)), dark=True, class_=class_)
                    v.BtnWithClick(children="+/-", on_click=lambda: dispatch(("negate", None)), dark=True, class_=class_)
                    v.BtnWithClick(children="%", on_click=lambda: dispatch(("percent", None)), dark=True, class_=class_)
                    v.BtnWithClick(children="/", color="primary", on_click=lambda: dispatch(("operator", operator_map["/"])), class_=class_)

                column_op = ["x", "-", "+"]
                for i in range(3):
                    with sol.HBox(grow=False):
                        for j in range(3):
                            digit = str(j + (2 - i) * 3 + 1)
                            v.BtnWithClick(children=digit, on_click=lambda digit=digit: dispatch(("digit", digit)), class_=class_)
                        op_symbol = column_op[i]
                        op = operator_map[op_symbol]
                        v.BtnWithClick(children=op_symbol, color="primary", on_click=lambda op=op: dispatch(("operator", op)), class_=class_)
                with sol.HBox(grow=False):
                    # v.Btn(children='gap', style_="visibility: hidden")
                    def boom():
                        print("boom")
                        raise ValueError("lala")

                    v.BtnWithClick(children="?", on_click=boom, class_=class_)

                    v.BtnWithClick(children="0", on_click=lambda: dispatch(("digit", "0")), class_=class_)
                    v.BtnWithClick(children=".", on_click=lambda: dispatch(("digit", ".")), class_=class_)

                    v.BtnWithClick(children="=", color="primary", on_click=lambda: dispatch(("calculate", None)), class_=class_)

    return main


App = Calculator
app = Calculator()



# import logging
# import sys

# import pytest

# from solara.server import reload
# from solara.server.app import AppScript

# logger = logging.getLogger("solara.server.app_test")


# def test_watch_module_reload(tmpdir, app_context, extra_include_path):
#     import ipyvuetify as v

#     with extra_include_path(str(tmpdir)):
#         py_file = tmpdir / "test.py"
#         py_mod_file = tmpdir / "somemod.py"

#         logger.info("writing files")
#         with open(py_mod_file, "w") as f:
#             f.write("import ipyvuetify as v; App = v.Btn.element\n")
#         with open(py_file, "w") as f:
#             f.write("import somemod; app=somemod.App\n")

#         logger.info("wrote files")

#         app = AppScript(f"{py_file}")
#         try:
#             result = app.run()
#             assert "somemod" in sys.modules
#             assert "somemod" in reload.reloader.watched_modules
#             assert result().component.widget == v.Btn

#             # change depending module
#             with open(py_mod_file, "w") as f:
#                 f.write("import ipyvuetify as v; App = v.Card.element\n")
#             # wait for the event to trigger
#             reload.reloader.reload_event_next.wait()
#             # assert "somemod" not in sys.modules
#             result = app.run()
#             assert "somemod" in sys.modules
#             assert result().component.widget == v.Card
#         finally:
#             app.close()
#             del sys.modules["somemod"]
#             reload.reloader.watched_modules.remove("somemod")


# def test_watch_module_import_error(tmpdir, app_context, extra_include_path):
#     import ipyvuetify as v

#     with extra_include_path(str(tmpdir)):
#         py_file = tmpdir / "test.py"
#         py_mod_file = tmpdir / "somemod2.py"

#         logger.info("writing files")
#         with open(py_mod_file, "w") as f:
#             f.write("import ipyvuetify as v; App = v.Btn.element\n")
#         with open(py_file, "w") as f:
#             f.write("import somemod2; app=somemod2.App\n")

#         logger.info("wrote files")

#         app = AppScript(f"{py_file}")
#         try:
#             # import pdb

#             # pdb.set_trace()
#             result = app.run()
#             assert "somemod2" in sys.modules
#             assert "somemod2" in reload.reloader.watched_modules
#             assert result().component.widget == v.Btn

#             # syntax error
#             with open(py_mod_file, "w") as f:
#                 f.write("import ipyvuetify as v; App !%#$@= v.Card.element\n")
#             reload.reloader.reload_event_next.wait()
#             # # assert "somemod" not in sys.modules
#             with pytest.raises(SyntaxError):
#                 result = app.run()

#             with open(py_mod_file, "w") as f:
#                 f.write("import ipyvuetify as v; App = v.Card.element\n")
#             reload.reloader.reload_event_next.wait()
#             result = app.run()
#             assert "somemod2" in sys.modules
#             assert result().component.widget == v.Card
#         finally:
#             app.close()
#             del sys.modules["somemod2"]
#             reload.reloader.watched_modules.remove("somemod2")



from solara.kitchensink import react, v


@react.component
def ColorCard(title, color):
    with v.Card(style_=f"background-color: {color}; width: 100%; height: 100%") as main:
        v.CardTitle(children=[title])
    return main



import pandas as pd
import plotly.express as px

from solara.kitchensink import react, sol

try:
    df = pd.read_csv("https://raw.githubusercontent.com/plotly/datasets/master/gapminderDataFiveYear.csv")
except Exception:
    df = None

if df is not None:
    year_min = df["year"].min()
    year_max = df["year"].max()
    years = df["year"].unique().tolist()


@react.component
def Plotly():
    with sol.Div() as main:
        sol.Markdown(
            """# Plotly
Solara supports plotly and plotly express. Create your figure (not a figure widget)
and pass it to the FigurePlotly component.
"""
        )
        index = sol.ui_slider(value=0, min=0, max=len(years) - 1, tick_labels=years, key="year slider index")
        selected_year = years[index]

        filtered_df = df[df.year == selected_year].copy()

        fig = px.scatter(filtered_df, x="gdpPercap", y="lifeExp", size="pop", color="continent", hover_name="country", log_x=True, size_max=55)
        fig.update_layout(transition_duration=1500)

        sol.FigurePlotly(fig, dependencies=[selected_year])
    return main


app = Plotly()



from pathlib import Path
from typing import cast

from solara.kitchensink import react, sol, v

from .api import API
from .demo import Demo
from .doc import Doc
from .libraries import Libraries

directory = Path(__file__).parent

md = open(directory / "README.md").read()


@react.component
def Home():
    with sol.Div(style_="padding: 20px") as main:
        sol.Markdown(md)
    return main


routes = [
    dict(component=Home, path="/", label="What is Solara ☀️?"),
    dict(component=Demo, path="/demo", label="Demo"),
    dict(component=API, path="/api", label="API"),
    dict(component=Libraries, path="/libraries", label="Supported libraries"),
    dict(component=Doc, path="/doc", label="Docs"),
]


@react.component
def All():
    tab, set_tab = react.use_state(0, "tab")

    route_current = routes[tab]
    path = route_current["path"]

    def on_location(location: str):
        matching_tabs = [tab for tab, route in enumerate(routes) if route["path"] == location]
        if matching_tabs:
            set_tab(matching_tabs[0])

    with v.Tabs(v_model=tab, on_v_model=set_tab) as main:
        sol.Navigator(location=path, on_location=on_location)

        for route in routes:
            with v.Tab(children=[route["label"]]):
                pass
        with v.TabsItems(v_model=tab):
            component = cast(react.core.Component, route_current["component"])
            key = key = route_current["path"]
            component().key(f"path_{key}")

    return main


app = All()



import threading
import time
from pathlib import Path
from typing import Optional, cast

from solara.kitchensink import react, sol, v, w

HERE = Path(__file__).parent
__doc__ = open(HERE / "use_thread.md").read()


@react.component
def UseThreadDemo():
    number, set_number = react.use_state(17)
    # the number that proofs it is not a prime
    proof, set_proof = react.use_state(cast(Optional[int], None))

    def work(cancelled: threading.Event):
        for i in range(3, number):
            reminder = number % i
            if reminder == 0:
                set_proof(i)
                return False
            # make it always take ~4 seconds
            time.sleep(4 / number)
        return True

    # work will be cancelled/restarted every time the dependency changes
    result: sol.Result[bool] = sol.use_thread(work, dependencies=[number])

    with sol.VBox() as main:
        w.IntText(value=number, on_value=set_number)
        if result.state == sol.ResultState.FINISHED:
            if result.value:
                sol.Success(f"{number} is a prime!")
            else:
                sol.Error(f"{number} is not a prime, it can be divided by {proof} ")
        elif result.state == sol.ResultState.ERROR:
            sol.Error(f"Error occurred: {result.error}")
        else:
            sol.Info(f"Running... (status = {result.state})")
            v.ProgressLinear(indeterminate=True)
    return main


App = UseThreadDemo



from typing import cast

from solara.kitchensink import react, v

from .altair import Altair
from .bqplot import Plot
from .plotly import Plotly

tabs = {
    "Bqplot": Plot,
    "Plotly": Plotly,
    "Altair": Altair,
}


@react.component
def Libraries():
    tab, set_tab = react.use_state(0, "tab")

    # md, set_md = use_state("")
    with v.Tabs(v_model=tab, on_v_model=set_tab, vertical=True) as main:
        for key in tabs:
            with v.Tab(children=[key]):
                pass
        component = cast(react.core.Component, list(tabs.values())[tab])
        with v.TabsItems(v_model=tab):
            component().key(str(tab))
    return main


app = Libraries()



import ipyvuetify as vy
import react_ipywidgets as react


class CodeHighlightCssWidget(vy.VuetifyTemplate):
    template_file = (__file__, "code_highlight_css.vue")


@react.component
def CodeHighlightCss():
    return CodeHighlightCssWidget.element()



import contextlib
import logging
import os
import sys
import threading
import time
from http.server import HTTPServer
from typing import Optional, Set, Union

import playwright.sync_api
import pytest
import uvicorn.server

import solara.server.app
import solara.server.server
import solara.server.settings
from solara.server import reload
from solara.server.starlette import app as app_starlette

logger = logging.getLogger("solara-test.integration")

worker = os.environ.get("PYTEST_XDIST_WORKER", "gw0")
TEST_PORT = 18765 + 2 + int(worker[2:])
SERVER = os.environ.get("SOLARA_SERVER")
if SERVER:
    SERVERS = [SERVER]
else:
    SERVERS = ["flask", "starlette"]


urls: Set[str] = set()


# allow symlinks on solara+starlette
solara.server.settings.main.mode = "development"


@pytest.fixture(scope="session")
def url_checks():
    yield
    non_localhost = [url for url in urls if not url.startswith("http://localhost")]
    allow_list = [
        "https://user-images.githubusercontent.com",
        "https://cdnjs.cloudflare.com/ajax/libs/emojione/2.2.7",
    ]
    non_allow_urls = [url for url in non_localhost if not any(url.startswith(allow) for allow in allow_list)]
    if non_allow_urls:
        msg = "The following URLs were not allowed (non local host, and not in allow list):\n"
        msg += "\n".join(non_allow_urls)
        raise AssertionError(msg)


# see https://github.com/microsoft/playwright-pytest/issues/23
@pytest.fixture
def context(context: playwright.sync_api.BrowserContext, url_checks):
    context.set_default_timeout(300000)

    def handle(route, request: playwright.sync_api.Request):
        urls.add(request.url)
        route.continue_()

    context.route("**/*", handle)
    yield context


class ServerBase(threading.Thread):
    def __init__(self, port: int, host: str = "localhost", **kwargs):
        self.port = port
        self.host = host
        self.base_url = f"http://{self.host}:{self.port}"

        self.kwargs = kwargs
        self.started = threading.Event()
        self.stopped = threading.Event()
        self.error: Optional[BaseException] = None
        self.server = None
        super().__init__(name="test-server-thread")
        self.setDaemon(True)

    def run(self):
        try:
            logger.info("Starting main loop")
            self.mainloop()
        except BaseException as e:  # noqa
            self.error = e
            self.started.set()
            logger.exception("Issue starting server")

    def serve_threaded(self):
        logger.debug("start thread")
        self.start()
        logger.debug("wait for thread to run")
        self.started.wait()
        if self.error:
            raise self.error

    def wait_until_serving(self):
        for n in range(30):
            if self.has_started():
                time.sleep(0.1)  # give some time to really start
                return
            time.sleep(0.05)
        else:
            raise RuntimeError(f"Server at {self.base_url} does not seem to be running")

    def serve(self):
        raise NotImplementedError

    def mainloop(self):
        logger.info("serving at http://%s:%d" % (self.host, self.port))
        try:
            self.serve()
        except:  # noqa: E722
            logger.exception("Oops, server stopped unexpectedly")
        finally:
            self.stopped.set()

    def stop_serving(self):
        logger.debug("stopping server")
        self.signal_stop()
        self.stopped.wait(10)
        if not self.stopped.is_set():
            logger.error("stopping server failed")
        else:
            logger.debug("stopped server")

    def signal_stop(self):
        pass

    def has_started(self) -> bool:
        return False


class ServerStarlette(ServerBase):
    server: uvicorn.server.Server

    def has_started(self):
        return self.server.started

    def signal_stop(self):
        self.server.should_exit = True
        # this cause uvicorn to not wait for background tasks, e.g.:
        # <Task pending name='Task-55'
        #  coro=<WebSocketProtocol.run_asgi() running at
        #  /.../uvicorn/protocols/websockets/websockets_impl.py:184>
        # wait_for=<Future pending cb=[<TaskWakeupMethWrapper object at 0x16896aa00>()]>
        # cb=[WebSocketProtocol.on_task_complete()]>
        self.server.force_exit = True
        self.server.lifespan.should_exit = True

    def serve(self):

        from uvicorn.config import Config
        from uvicorn.server import Server

        if sys.version_info[:2] < (3, 7):
            # make python 3.6 work
            import asyncio

            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

        # uvloop will trigger a: RuntimeError: There is no current event loop in thread 'fastapi-thread'
        config = Config(app_starlette, host=self.host, port=self.port, **self.kwargs, loop="asyncio")
        self.server = Server(config=config)
        self.started.set()
        self.server.run()


def _serve_flask_in_process(port, host):
    from solara.server.flask import app

    app.run(debug=False, port=port, host=host)


class ServerFlask(ServerBase):
    def has_started(self):
        import socket

        s = socket.socket()
        try:
            s.connect((self.host, self.port))
        except ConnectionRefusedError:
            return False
        return True

    def signal_stop(self):
        assert isinstance(self.server, HTTPServer)
        self.server.shutdown()  # type: ignore

    def serve(self):
        from werkzeug.serving import make_server

        from solara.server.flask import app

        self.server = make_server(self.host, self.port, app, threaded=True)  # type: ignore
        assert isinstance(self.server, HTTPServer)
        self.started.set()
        self.server.serve_forever(poll_interval=0.05)  # type: ignore


server_classes = {
    "flask": ServerFlask,
    "starlette": ServerStarlette,
}


@pytest.fixture(params=SERVERS)
def solara_server(request):
    server_class = server_classes[request.param]
    global TEST_PORT
    webserver = server_class(TEST_PORT)

    try:
        webserver.serve_threaded()
        webserver.wait_until_serving()
        yield webserver
    finally:
        webserver.stop_serving()
        TEST_PORT += 1


@pytest.fixture()
def solara_app(solara_server):
    @contextlib.contextmanager
    def run(app: Union[solara.server.app.AppScript, str]):
        solara.server.server.solara_app.close()
        if isinstance(app, str):
            app = solara.server.app.AppScript(app)
        solara.server.server.solara_app = app
        try:
            yield
        finally:
            if app.type == solara.server.app.AppType.MODULE:
                del sys.modules[app.name]
                reload.reloader.watched_modules.remove(app.name)

            app.close()

    return run



import altair as alt
import pandas as pd
from vega_datasets import data

from solara.kitchensink import react, sol


@react.component
def Altair():
    click_data, set_click_data = react.use_state(None)
    hover_data, set_hover_data = react.use_state(None)

    if 0:
        # to bad this example doesn't work well with on_click and on_hover
        source = data.cars()
        chart = (
            alt.Chart(source)
            .mark_circle(size=60)
            .encode(x="Horsepower", y="Miles_per_Gallon", color="Origin", tooltip=["Name", "Origin", "Horsepower", "Miles_per_Gallon"])
            .interactive()
        )
    else:
        source = pd.DataFrame({"a": ["A", "B", "C", "D", "E", "F", "G", "H", "I"], "b": [28, 55, 43, 91, 81, 53, 19, 87, 52]})

        chart = alt.Chart(source).mark_bar().encode(x="a", y="b")

    with sol.Div() as main:
        sol.MarkdownIt(
            """
Altair is supported since we can render vega lite.
We also support on_click and on_hover events.
```python
source = pd.DataFrame({{
    "a": ["A", "B", "C", "D", "E", "F", "G", "H", "I"],
    "b": [28, 55, 43, 91, 81, 53, 19, 87, 52]
}})
chart = alt.Chart(source).mark_bar().encode(x="a", y="b")
sol.AltairChart(chart, on_click=set_click_data, on_hover=set_hover_data)
```
            """
        )
        sol.AltairChart(chart, on_click=set_click_data, on_hover=set_hover_data)

        sol.Markdown(
            f"""
Click data:

```
{click_data}
```

Hover data:
```
{hover_data}
```
            """
        )

    return main


app = Altair()



import react_ipywidgets as react  # noqa: F401
from react_ipywidgets import ipyvue as vue  # noqa: F401
from react_ipywidgets import ipyvuetify as v  # noqa: F401
from react_ipywidgets import ipywidgets as w  # noqa: F401

import solara as sol  # noqa: F401

from . import layout  # noqa: F401



"""
# HTML

Insert a custom html tag, possible with unescaped html text inside.

Note that this will be interpreted by the browser, so make sure the input html text
cannot include code from users.
"""

from solara.kitchensink import react, sol


@react.component
def HTMLDemo():
    html = """
<h1>Custom html</h1>
<ul>
    <li>Item 1
    <li>Item 2
</ul>
"""
    with sol.VBox() as main:
        sol.HTML(tag="div", unsafe_innerHTML=html)
    return main


Component = sol.HTML
App = HTMLDemo
